import logging
import logging.handlers as _handlers
import os as _os

from simbak import config as _config


def set_logger():
    stream_handler = _stream_handler()
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(stream_handler)
    return root_logger


def set_file_logger(log_path: str):
    root_logger = logging.getLogger()

    if log_path is not None:
        _set_log_path(log_path)
        rotating_file_handler = _rotating_file_handler(log_path)
        root_logger.addHandler(rotating_file_handler)

    return root_logger


def _stream_handler():
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler = logging.StreamHandler()

    if _config.SIMBAK_ENV == 'debug':
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.INFO)

    handler.setFormatter(formatter)
    return handler


def _set_log_path(log_path: str) -> str:
    if _os.path.exists(log_path) is False:
        _os.mkdir(log_path)
    return log_path


def _rotating_file_handler(log_path: str):
    formatter = logging.Formatter(
        fmt='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%Y-%m-%d, %H-%M-%S')
    handler = _handlers.RotatingFileHandler(
        _os.path.join(log_path, 'simbak.log'),
        maxBytes=1000000,
        backupCount=20)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    return handler



from simbak.agent.base import BaseAgent
from simbak.agent.normal import NormalAgent
from simbak.agent.rotating import RotatingAgent

__all__ = [
    'BaseAgent',
    'NormalAgent',
    'RotatingAgent',
]



import logging as _logging

from simbak.agent.base import BaseAgent as _BaseAgent

_logger = _logging.getLogger(__name__)


class NormalAgent(_BaseAgent):
    def __init__(self, sources: list, destinations: list, name: str,
                 compression_level: int = 6):
        """
        Args:
            sources (list of str): Paths to the files that you are
                backing up.
            destinations (list of str): Paths of where you want the
                backup to be stored.
            name (str): Name of the backup, this will name the backup
                files.
            compression_level (int, optional): The gzip compression
                level that you want to use for the backup. Default to 6.
        """
        super().__init__(sources, destinations, name, compression_level)

    def backup(self):
        """Standard simbak backup.

        This will backup all the files defined in the sources and store
        them in a gzip'd file in each of the destinations. The name of
        the gzip file will be the name parameter suffixed with a time
        stamp, the format of the timestamp is YYYY-MM-DD--hh-mm-ss
        """
        _logger.info(f'Starting backup [{self._name}]')
        super()._backup()



import abc as _abc
import logging as _logging
import os as _os

from simbak import fileutil as _fileutil
from simbak.exception import BackupError as _BackupError
from simbak.fileutil import dir_size as _dir_size

_logger = _logging.getLogger(__name__)


class BaseAgent(_abc.ABC):
    def __init__(self, sources: list, destinations: list, name: str,
                 compression_level: int):
        """Default initializer for all agent objects.

        Args:
            sources (list of str): Paths to the files that you are
                backing up.
            destinations (list of str): Paths of where you want the
                backup to be stored.
            name (str): Name of the backup, this will name the backup
                files.
            compression_level (int, optional): The gzip compression
                level that you want to use for the backup.
        """
        self._sources = sources
        self._destinations = destinations
        self._name = name
        self._compression_level = compression_level

        self.__filter_paths()

    @_abc.abstractmethod
    def backup(self):
        pass

    def __filter_paths(self, create: bool = True):
        self._sources = _fileutil.filter_paths(self._sources)
        self._destinations = _fileutil.filter_paths(
            self._destinations, create=create)
        self._validate_paths()

    def _validate_paths(self):
        if len(self._sources) == 0:
            raise _BackupError('No sources for backup')
        if len(self._destinations) == 0:
            raise _BackupError('No destinations for backup')

    def _sources_size(self):
        total_size = 0
        for source in self._sources:
            dir_size = _dir_size(source)
            total_size += dir_size

        return total_size

    def _backup(self):
        _logger.info(
            f'{self._sources_size()} bytes for [{self._name}] sources')

        # A unique file name is important for certain backup agents.
        file_name = _fileutil.unique_file_name(self._name)
        _logger.info(f'Backup file name will be {file_name}')

        # Using tar gzip for the backup. Tarring once, distributing later.
        first_path = _fileutil.create_targz(
            sources=self._sources,
            destination=self._destinations[0],
            file_name=file_name,
            compression_level=self._compression_level)
        backup_size = _os.path.getsize(first_path)
        _logger.info(f'{backup_size} bytes for [{self._name}] backup file')

        _fileutil.distribute_file(
            path=first_path,
            destinations=self._destinations[1:])



import logging as _logging
import os as _os
import shutil as _shutil
import tarfile as _tarfile
from datetime import datetime as _datetime
from typing import Optional as _Optional

_logger = _logging.getLogger(__name__)


def filter_paths(paths: list, create=False) -> list:
    """Gets only existing paths and cleans the paths to be OS compliant.

    Args:
        paths (list of str): The list of paths that needs to be
            filtered.
        create (bool, optional): Create a path if it doesn't exist.
            Defaults to False.

    Returns:
        list of str: The filtered list of paths.
    """
    filtered_paths = []

    for path in paths:
        normpath = _os.path.normpath(path)

        if _os.path.exists(normpath):
            filtered_paths.append(normpath)
        elif create:
            _logger.info(
                f'{normpath} doesn\'t exist, creating directory '
                'with that path.')
            _os.makedirs(normpath)
            filtered_paths.append(normpath)
        else:
            _logger.warning(
                f'Failed to access {normpath}, it doesn\'t exist')

    return filtered_paths


def unique_file_name(name: str) -> str:
    """Creates a unique file name by suffixing date and time.

    Args:
        name (str): The prefix of the unique file name.

    Returns:
        str: The unique file in the format of
            <name>--YYYY-MM-DD--hh-mm-ss.tar.gz
    """
    time_suffix = _datetime.now().strftime('%Y-%m-%d--%H-%M-%S')
    file_name = f'{name}--{time_suffix}.tar.gz'
    return file_name


def distribute_file(path: str, destinations: list):
    """Distributes a backup file to other destinations.

    Args:
        path (str): Path to the file that you want to distribute.
        destinations (list of str): Paths of where you want the file to
            be stored.
    """
    file_name = _os.path.basename(path)

    for destination in destinations:
        to_path = _os.path.join(destination, file_name)
        _shutil.copyfile(path, to_path)
        _logger.info(f'Saved backup {file_name} to {destination}')


def create_targz(sources: list, destination: str, file_name: str,
                 compression_level: int = 6) -> str:
    """Creates a tar.gz file in a single destination.

    If you try to create a tar.gz file for a file that already exists,
    it will not overwrite that file.

    Args:
        sources (list of str): The paths of the files you want to file.
        destination (str): The path of where you want the file to be
            stored.
        file_name (str): The name that the file will be.
        compression_level (int): The gzip compression level that you
            want to use for the file.

    Returns:
        str: Path of the file that was created, if the file already
            existed, it will return the already existing file path.
    """
    path = _os.path.join(destination, file_name)

    try:
        tar = _tarfile.open(
            path, 'x:gz', compresslevel=compression_level)

        for source in sources:
            basename = _os.path.basename(source)
            _logger.info(f'Compressing {source}')
            try:
                tar.add(source, basename)
            except PermissionError:
                _logger.error(
                    f'Couldn\'t compress {source}, pemission denied.')

        tar.close()
        _logger.info(f'Saved backup {file_name} to {destination}')
    except FileExistsError:
        _logger.error(
            f'Failed to create backup {file_name}, file already exists')

    return path


def dir_size(base_path: str = '.') -> int:
    """Gets the total size of a directory, in bytes.

    Args:
        base_path (str, optional): Path to the dir to get the size of.

    Returns:
        int: Size of the directory, in bytes.
    """
    size = 0

    for dir_path, dir_names, file_names in _os.walk(base_path):
        for file_name in file_names:
            file_path = _os.path.join(dir_path, file_name)

            # Skip if it is symbolic link.
            if not _os.path.islink(file_path):
                size += _os.path.getsize(file_path)

    return size


def oldest_file(file_names: list) -> _Optional[str]:
    """Gets the oldest file from a list of backup file names.

    Args:
        file_names (list of str): List of backup file names.

    Returns:
        str, optional: Oldest file name.
    """
    if len(file_names) == 0:
        return None

    oldest = file_names[0]
    oldest_time = _time_from_file_name(file_names[0])

    for file_name in file_names:
        time = _time_from_file_name(file_name)

        if time < oldest_time:
            oldest = file_name
            oldest_time = time

    return oldest


def _time_from_file_name(file_name: str):
    # Remove the .tar.gz from file name
    file_name = file_name[:-7]

    parts = file_name.split('--')
    time_str = f'{parts[1]} {parts[2]}'
    return _datetime.strptime(time_str, '%Y-%m-%d %H-%M-%S')



import logging as _logging
import os as _os

from simbak import fileutil as _fileutil
from simbak.agent.base import BaseAgent as _BaseAgent

_logger = _logging.getLogger(__name__)


class RotatingAgent(_BaseAgent):
    def __init__(self, sources: list, destinations: list, name: str,
                 rotate_limit: int, compression_level: int = 6):
        """
        Args:
            sources (list of str): Paths to the files that you are
                backing up.
            destinations (list of str): Paths of where you want the
                backup to be stored.
            name (str): Name of the backup, this will name the backup
                files.
            rotate_limit (int): The maximum amount of backups to keep in
                a destination.
            compression_level (int, optional): The gzip compression
                level that you want to use for the backup. Default to 6.
        """
        super().__init__(sources, destinations, name, compression_level)
        self._rotate_limit = rotate_limit

    def backup(self):
        """Rotating simbak backup.

        This will backup all the files defined in the sources and store
        them in a gzip'd file in each of the destinations. The name of
        the gzip file will be the `_name` property suffixed with a time
        stamp, the format of the timestamp is YYYY-MM-DD--hh-mm-ss

        The amount of backups in the destination will be limited by the
        `_rotate_limit` value, the oldest backup will be removed if the
        limit is exceeded.
        """
        _logger.info(f'Starting rotating backup [{self._name}]')
        _logger.info(f'Rotating limit is {self._rotate_limit}')
        super()._backup()

        for destination in self._destinations:
            # Get the valid files for this rotating agent.
            all_files = _os.listdir(destination)
            targz_files = [f for f in all_files if f.endswith('.tar.gz')]
            valid_files = [f for f in targz_files if f.startswith(self._name)]

            # Remove the oldest file(s) if the limit is surpassed.
            if len(valid_files) > self._rotate_limit:
                _logger.info(
                    f'Rotate limit has been exceeded for [{self._name}] in '
                    f'{destination}')
                oldest_file_name = _fileutil.oldest_file(valid_files)
                oldest_file_path = _os.path.join(destination, oldest_file_name)
                _logger.info(f'Removing oldest backup {oldest_file_path}')
                try:
                    _os.remove(oldest_file_path)
                except PermissionError:
                    _logger.error(
                        f'Couldn\'t remove {oldest_file_path}, permission '
                        f'denied')



import logging as _logging
import sys

from simbak.cli import main
from simbak.exception import BackupError

_logger = _logging.getLogger(__name__)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
    except BackupError as backup_error:
        _logger.error(backup_error)
        sys.exit(1)



from dotenv import load_dotenv as _load_dotenv
from simbak.logging import set_logger as _set_logger

__version__ = '0.4.1'

_load_dotenv()
_set_logger()


def backup(sources: list, destinations: list, name: str,
           compression_level: int = 6, log_path: str = None):
    """The easiest way to perform a standard backup.

    Args:
        sources (list of str): Paths to the files that you are backing
            up.
        destinations (list of str): Paths of where you want the backup
            to be stored.
        name (str): Name of the backup, this will name the backup files.
        compression_level (int, optional): The gzip compression level
            that you want to use for the backup. Defaults to 6.
        log_path (str, optional): The file location to store the logs.
    """
    from simbak.agent.normal import NormalAgent
    from simbak.logging import set_file_logger

    if log_path is not None:
        set_file_logger(log_path)

    agent = NormalAgent(sources, destinations, name, compression_level)
    agent.backup()
